from typing import Literal, Optional, cast

from qdrant_client.conversions.common_types import Filter

from ..schemas import QueryBase


def cleanse_query_for_search(
    query: Optional[QueryBase | dict], offset_type: Literal["int", "str"] = "int"
) -> Filter | None:
    if query is None:
        return None
    elif isinstance(query, dict):
        query = QueryBase(**query)
    if offset_type == "int":
        assert query.offset is None or isinstance(query.offset, int), (
            "[Db] Offset must be int in Qdrant searching"
        )
    elif offset_type == "str":
        assert query.offset is None or isinstance(query.offset, str), (
            "[Db] Offset must be id (str) in Qdrant querying"
        )
    assert query.id is None, (
        "[Db] QueryBase's 'id' key should not being use for similarity search in qdrant"
    )
    assert query.ids is None, (
        "[Db] QueryBase's 'ids' key should not being use for similarity search in qdrant"
    )
    assert query.order_by is None, (
        "[Db] QueryBase's 'order_by' key should not being use for similarity search in qdrant"
    )
    assert query.filters is None or isinstance(query.filters, dict), (
        "[Db] Query.filters only accept dict of qdrant query"
    )

    base = query.filters if query.filters is not None else {}

    # forbidden id and ids in filters
    if "id" in base:
        raise AssertionError(
            "[Db] You should not include 'id' in your QueryBase.filters when searching Qdrant"
        )
    if "ids" in base:
        raise AssertionError(
            "[Db] You should not include 'ids' in your QueryBase.filters when searching Qdrant"
        )
    return cast(Filter, base)



from typing import Generator, Optional

from motor.motor_asyncio import AsyncIOMotorClient
from pymongo import MongoClient
from qdrant_client import AsyncQdrantClient, QdrantClient
from redis import Redis
from redis.asyncio import Redis as AsyncRedis
from sqlmodel import Session

from keble_db.wrapper import ExtendedAsyncRedis
from keble_db.wrapper.redis import ExtendedRedis

from ..session import Db


class ApiDbDeps:
    def __init__(self, db: Db):
        self.__db = db

    def get_redis(self) -> Generator[Redis | None, None, None]:
        r = None
        try:
            r = self.__db.get_redis(force_new_instance=True)
            yield r
        finally:
            self.__db.try_close(r)

    def get_aredis(self) -> Generator[AsyncRedis | None, None, None]:
        r = None
        try:
            r = self.__db.get_aredis(new_instance=True)
            yield r
        finally:
            # skip async closing
            # this redis instance is using connection pool which you do not need to close it manually.
            pass
            # self.__db.try_close(r)

    def get_extended_aredis(
        self, *, namespace: Optional[str] = None
    ) -> Generator[ExtendedAsyncRedis | None, None, None]:
        r = None
        try:
            r = ExtendedAsyncRedis.extend(
                aredis=self.__db.get_aredis(new_instance=True), namespace=namespace
            )
            yield r
        finally:
            # skip async closing
            # this redis instance is using connection pool which you do not need to close it manually.
            pass
            # self.__db.try_close(r)

    def get_extended_redis(
        self, *, namespace: Optional[str] = None
    ) -> Generator[ExtendedRedis | None, None, None]:
        r = None
        try:
            r = ExtendedRedis.extend(redis=self.__db.get_redis(), namespace=namespace)
            yield r
        finally:
            pass

    def get_mongo(self) -> Generator[MongoClient | None, None, None]:
        m = None
        try:
            m = self.__db.get_mongo(new_instance=True)
            yield m
        finally:
            self.__db.try_close(m)

    def get_amongo(self) -> Generator[AsyncIOMotorClient | None, None, None]:
        m = None
        try:
            m = self.__db.get_amongo()
            yield m
        finally:
            self.__db.try_close(m)

    def get_write_sql(self) -> Generator[Session | None, None, None]:
        s = None
        try:
            s = self.__db.get_sql_write_client(new_instance=True)
            yield s
        finally:
            self.__db.try_close(s)

    def get_qdrant(self) -> Generator[QdrantClient | None, None, None]:
        q = None
        try:
            q = self.__db.get_qdrant_client(new_instance=True)
            yield q
        finally:
            self.__db.try_close(q)

    def get_aqdrant(self) -> Generator[AsyncQdrantClient | None, None, None]:
        q = None
        try:
            q = self.__db.get_aqdrant_client(new_instance=True)
            yield q
        finally:
            self.__db.try_close(q)



from typing import Any, Generic, List, Optional, Sequence, Tuple, Type, TypeVar, cast

from pydantic import BaseModel
from qdrant_client import AsyncQdrantClient, QdrantClient
from qdrant_client import models as qdrant_models
from qdrant_client.conversions import common_types as qdrant_types
from qdrant_client.models import UpdateResult

from ..schemas import QueryBase
from .qdrant_util import cleanse_query_for_search

ModelType = TypeVar("ModelType", bound=BaseModel)
VectorModelType = TypeVar("VectorModelType", bound=BaseModel)


class Record(qdrant_types.Record, Generic[VectorModelType]):
    payload: Optional[Any] = None
    score: Optional[float] = None
    vector: Optional[VectorModelType] = None


class QdrantCRUDBase(Generic[ModelType, VectorModelType]):
    def __init__(
        self,
        model: Type[ModelType],
        vector_model: Type[VectorModelType],
        collection: str,
    ):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD) in Qdrant

        **Parameters**
        * `model`: A Qdrant Alchemy model class
        """
        self.model: Type[ModelType] = model
        self.vector_model: Type[VectorModelType] = vector_model
        self.collection = collection

    def __is_result_success(self, result: qdrant_types.UpdateResult):
        return (
            result.status.lower() == qdrant_models.UpdateStatus.COMPLETED.value.lower()
        )

    def __convert_record(self, record: qdrant_types.Record) -> Record[VectorModelType]:
        assert record.vector is None or isinstance(record.vector, dict), (
            f"[Db] Expected record.vector to be a dict or None, but got {record.vector}"
        )
        return Record(
            **record.model_dump(exclude={"payload", "vector"}),
            payload=self.model(**record.payload)
            if record.payload is not None
            else None,
            vector=self.vector_model(**record.vector)
            if record.vector is not None
            else None,
        )

    def __convert_records(self, records: List[qdrant_types.Record]) -> List[Record]:
        return [self.__convert_record(record=record) for record in records]

    def get_multi_records_by_ids(
        self, client: QdrantClient, *, _ids: Sequence[qdrant_types.PointId], **kwargs
    ) -> List[Record]:
        return self.__convert_records(
            client.retrieve(self.collection, ids=_ids, **kwargs)
        )

    async def aget_multi_records_by_ids(
        self,
        client: AsyncQdrantClient,
        *,
        _ids: Sequence[qdrant_types.PointId],
        **kwargs,
    ) -> List[Record]:
        return self.__convert_records(
            await client.retrieve(self.collection, ids=_ids, **kwargs)
        )

    def get_multi_by_ids(
        self, client: QdrantClient, *, _ids: Sequence[qdrant_types.PointId], **kwargs
    ) -> List[ModelType]:
        records = self.get_multi_records_by_ids(client, _ids=_ids, **kwargs)
        return [r.payload for r in records if r.payload is not None]

    async def aget_multi_by_ids(
        self,
        client: AsyncQdrantClient,
        *,
        _ids: Sequence[qdrant_types.PointId],
        **kwargs,
    ) -> List[ModelType]:
        records = await self.aget_multi_records_by_ids(client, _ids=_ids, **kwargs)
        return [r.payload for r in records if r.payload is not None]

    def first_record_by_id(
        self, client: QdrantClient, *, _id: qdrant_types.PointId, **kwargs
    ) -> Optional[Record]:
        records: List[Record] = self.get_multi_records_by_ids(
            client, _ids=[_id], **kwargs
        )
        if len(records) < 1:
            return None
        return records[0]

    async def afirst_record_by_id(
        self, client: AsyncQdrantClient, *, _id: qdrant_types.PointId, **kwargs
    ) -> Optional[Record]:
        records: List[Record] = await self.aget_multi_records_by_ids(
            client, _ids=[_id], **kwargs
        )
        if len(records) < 1:
            return None
        return records[0]

    def first_by_id(
        self, client: QdrantClient, *, _id: qdrant_types.PointId, **kwargs
    ) -> Optional[ModelType]:
        record = self.first_record_by_id(client, _id=_id, **kwargs)
        if record is not None:
            return record.payload

    async def afirst_by_id(
        self, client: AsyncQdrantClient, *, _id: qdrant_types.PointId, **kwargs
    ) -> Optional[ModelType]:
        record = await self.afirst_record_by_id(client, _id=_id, **kwargs)
        if record is not None:
            return record.payload

    @classmethod
    def _build_search_kwargs(cls, query: Optional[QueryBase | dict] = None):
        filter_ = cleanse_query_for_search(query, offset_type="int")
        offset = None
        if isinstance(query, dict):
            query = QueryBase(**query)
        if query is not None and query.offset is not None:
            assert isinstance(query.offset, int), (
                f"[Db] Expected search in qdrant with int or None type offset, but got {query.offset}"
            )
            offset = query.offset
        return {
            "query_filter": filter_,
            "limit": query.limit
            if query is not None and query.limit is not None
            else 100,  # by default 100
            "offset": offset,
        }

    def search(
        self,
        client: QdrantClient,
        *,
        vector: List[float],
        vector_key: str,
        query: Optional[QueryBase | dict] = None,
        score_threshold: float = 0.76,
        # 0.76 ~ 0.8 is a decent threshold base on our test, among different languages
        **kwargs,
    ) -> List[Record]:
        scored_points: List[qdrant_types.ScoredPoint] = client.search(
            self.collection,
            query_vector=(vector_key, vector),
            score_threshold=score_threshold,
            **self._build_search_kwargs(query),
            **kwargs,
        )
        return [
            Record(
                **sp.model_dump(include={"score", "shard_key", "id"}),
                payload=self.model(**sp.payload) if sp.payload is not None else None,
                vector=self.vector_model(**sp.vector)  # type: ignore
                if sp.vector is not None
                else None,
            )
            for sp in scored_points
        ]

    async def asearch(
        self,
        client: AsyncQdrantClient,
        *,
        vector: List[float],
        vector_key: str,
        query: Optional[QueryBase | dict] = None,
        score_threshold: float = 0.76,
        # 0.76 ~ 0.8 is a decent threshold base on our test, among different languages
        **kwargs,
    ) -> List[Record]:
        scored_points: List[qdrant_types.ScoredPoint] = await client.search(
            self.collection,
            query_vector=(vector_key, vector),
            score_threshold=score_threshold,
            **self._build_search_kwargs(query),
            **kwargs,
        )
        return [
            Record(
                **sp.model_dump(include={"score", "shard_key", "id"}),
                payload=self.model(**sp.payload) if sp.payload is not None else None,
                vector=self.vector_model(**sp.vector)  # type: ignore
                if sp.vector is not None
                else None,
            )
            for sp in scored_points
        ]

    @classmethod
    def _build_scroll_kwargs(cls, query: Optional[QueryBase | dict] = None):
        if isinstance(query, dict):
            query = QueryBase(**query)
        filter_ = cleanse_query_for_search(query, offset_type="str")
        assert query is None or query.limit is not None, (
            "[Db] You need to provide limit to perform a scroll in Qdrant"
        )
        offset = None
        if query is not None and query.offset is not None:
            assert isinstance(query.offset, str), (
                f"[Db] Expected str or None type query offset when scrolling records in qdrant, but got {query.offset}"
            )
            offset = query.offset

        return {
            "scroll_filter": filter_,
            "limit": query.limit
            if query is not None and query.limit is not None
            else 100,
            "offset": offset,
        }

    def scroll_records(
        self, client: QdrantClient, *, query: QueryBase | dict | None, **kwargs
    ) -> Tuple[List[Record], Optional[qdrant_types.PointId]]:
        records, next_point_id = client.scroll(
            self.collection, **self._build_scroll_kwargs(query), **kwargs
        )
        return self.__convert_records(records), next_point_id

    async def ascroll_records(
        self, client: AsyncQdrantClient, *, query: QueryBase | dict | None, **kwargs
    ) -> Tuple[List[Record], Optional[qdrant_types.PointId]]:
        records, next_point_id = await client.scroll(
            self.collection, **self._build_scroll_kwargs(query), **kwargs
        )
        return self.__convert_records(records), next_point_id

    def scroll(
        self, client: QdrantClient, *, query: QueryBase | dict | None, **kwargs
    ) -> Tuple[List[ModelType], Optional[qdrant_types.PointId]]:
        records, next_point_id = self.scroll_records(client, query=query, **kwargs)
        return [
            record.payload for record in records if record.payload is not None
        ], next_point_id

    async def ascroll(
        self, client: AsyncQdrantClient, *, query: QueryBase | dict | None, **kwargs
    ) -> Tuple[List[ModelType], Optional[qdrant_types.PointId]]:
        records, next_point_id = await self.ascroll_records(
            client, query=query, **kwargs
        )
        return [
            record.payload for record in records if record.payload is not None
        ], next_point_id

    def _return_create_response(self, response: UpdateResult):
        _, status = response
        status = status[1] if isinstance(status, tuple) and len(status) > 1 else status
        return str(status).lower() == qdrant_models.UpdateStatus.COMPLETED.value.lower()

    def create(
        self,
        client: QdrantClient,
        vector: VectorModelType,
        payload: ModelType,
        _id: qdrant_types.PointId,
        **kwargs,
    ) -> bool:
        vector_dict = vector.model_dump(exclude_none=True, exclude_unset=True)
        return self._return_create_response(
            client.upsert(
                collection_name=self.collection,
                points=[
                    qdrant_models.PointStruct(
                        id=cast(qdrant_models.ExtendedPointId, _id),
                        vector=vector_dict,
                        payload=payload.model_dump(),
                    )
                ],
                **kwargs,
            )
        )

    async def acreate(
        self,
        client: AsyncQdrantClient,
        vector: VectorModelType,
        payload: ModelType,
        _id: qdrant_types.PointId,
        **kwargs,
    ) -> bool:
        vector_dict = vector.model_dump(exclude_none=True, exclude_unset=True)
        return self._return_create_response(
            await client.upsert(
                collection_name=self.collection,
                points=[
                    qdrant_models.PointStruct(
                        id=cast(qdrant_models.ExtendedPointId, _id),
                        vector=vector_dict,
                        payload=payload.model_dump(),
                    )
                ],
                **kwargs,
            )
        )

    def create_multi(
        self,
        client: QdrantClient,
        *,
        payloads_and_vectors: List[Tuple[str, ModelType, VectorModelType]],
        **kwargs,
    ) -> bool:
        points: List[qdrant_models.PointStruct] = [
            qdrant_models.PointStruct(
                id=_id,
                vector=vector.model_dump(exclude_none=True, exclude_unset=True),
                payload=payload.model_dump(),
            )
            for _id, payload, vector in payloads_and_vectors
        ]
        return self._return_create_response(
            client.upsert(collection_name=self.collection, points=points, **kwargs)
        )

    async def acreate_multi(
        self,
        client: AsyncQdrantClient,
        *,
        payloads_and_vectors: List[Tuple[str, ModelType, VectorModelType]],
        **kwargs,
    ) -> bool:
        points: List[qdrant_models.PointStruct] = [
            qdrant_models.PointStruct(
                id=_id,
                vector=vector.model_dump(exclude_none=True, exclude_unset=True),
                payload=payload.model_dump(),
            )
            for _id, payload, vector in payloads_and_vectors
        ]
        return self._return_create_response(
            await client.upsert(
                collection_name=self.collection, points=points, **kwargs
            )
        )

    def delete(
        self, client: QdrantClient, *, _id: qdrant_types.PointId, **kwargs
    ) -> bool:
        update_result: qdrant_types.UpdateResult = client.delete(
            self.collection, points_selector=[_id], **kwargs
        )
        return self.__is_result_success(update_result)

    async def adelete(
        self, client: AsyncQdrantClient, *, _id: qdrant_types.PointId, **kwargs
    ) -> bool:
        update_result: qdrant_types.UpdateResult = await client.delete(
            self.collection, points_selector=[_id], **kwargs
        )
        return self.__is_result_success(update_result)

    def _get_points_selector_by_delete_query(
        self,
        query: QueryBase | dict | None,
    ) -> qdrant_models.PointIdsList | qdrant_models.FilterSelector:
        if query is None:
            query_dict: dict = {}
        else:
            if isinstance(query, dict):
                query = QueryBase(**query)
            query_dict: dict = query.model_dump(
                exclude_none=True, exclude_unset=True, exclude_defaults=True
            )
        filters_exist = "filters" in query_dict and query_dict["filters"] is not None
        ids_exist = "ids" in query_dict and query_dict["ids"] is not None
        if not (filters_exist or ids_exist):
            raise ValueError(
                "[Db] you must set query.ids or query.filters to delete_multi in qdrant"
            )
        if ids_exist:
            points_selector = qdrant_models.PointIdsList(points=query_dict["ids"])
        elif filters_exist:
            points_selector = qdrant_models.FilterSelector(filter=query_dict["filters"])
        else:
            raise Exception(
                "[Db] you must set query.ids or query.filters to delete_multi in qdrant"
            )
        return points_selector

    def delete_multi(
        self, client: QdrantClient, *, query: QueryBase | dict | None, **kwargs
    ):
        points_selector = self._get_points_selector_by_delete_query(query)
        update_result: qdrant_types.UpdateResult = client.delete(
            self.collection, points_selector=points_selector, **kwargs
        )
        return self.__is_result_success(update_result)

    async def adelete_multi(
        self, client: AsyncQdrantClient, *, query: QueryBase | dict | None, **kwargs
    ):
        points_selector = self._get_points_selector_by_delete_query(query)
        update_result: qdrant_types.UpdateResult = await client.delete(
            self.collection, points_selector=points_selector, **kwargs
        )
        return self.__is_result_success(update_result)

    def update_payload(
        self,
        client: QdrantClient,
        *,
        _id: qdrant_types.PointId,
        payload: ModelType,
        **kwargs,
    ):
        """Set Payload key value, not completely override entire payload object"""
        old_payload = self.first_by_id(client, _id=_id)
        if old_payload is not None:
            obj_dict = old_payload.model_dump()
            obj_dict.update(payload.model_dump())
            res = client.set_payload(
                collection_name=self.collection,
                payload=self.model(**obj_dict).model_dump(),
                points=[_id],
                **kwargs,
            )
            return self.__is_result_success(res)
        return False

    async def aupdate_payload(
        self,
        client: AsyncQdrantClient,
        *,
        _id: qdrant_types.PointId,
        payload: ModelType,
        **kwargs,
    ):
        """Set Payload key value, not completely override entire payload object"""
        old_payload = await self.afirst_by_id(client, _id=_id)
        if old_payload is not None:
            obj_dict = old_payload.model_dump()
            obj_dict.update(payload.model_dump())
            res = await client.set_payload(
                collection_name=self.collection,
                payload=self.model(**obj_dict).model_dump(),
                points=[_id],
                **kwargs,
            )
            return self.__is_result_success(res)
        return False

    def overwrite_payload(
        self,
        client: QdrantClient,
        *,
        _id: qdrant_types.PointId,
        payload: ModelType,
        **kwargs,
    ):
        """Completely override entire payload object, replace entire old payload"""
        res = client.overwrite_payload(
            collection_name=self.collection,
            payload=payload.model_dump(),
            points=[_id],
            **kwargs,
        )
        return self.__is_result_success(res)

    async def aoverwrite_payload(
        self,
        client: AsyncQdrantClient,
        *,
        _id: qdrant_types.PointId,
        payload: ModelType,
        **kwargs,
    ):
        """Completely override entire payload object, replace entire old payload"""
        res = await client.overwrite_payload(
            collection_name=self.collection,
            payload=payload.model_dump(),
            points=[_id],
            **kwargs,
        )
        return self.__is_result_success(res)

    def update_vector(
        self,
        client: QdrantClient,
        *,
        _id: qdrant_types.PointId,
        vector: VectorModelType,
        **kwargs,
    ) -> bool:
        res = client.update_vectors(
            collection_name=self.collection,
            points=[
                qdrant_models.PointVectors(
                    id=cast(qdrant_models.ExtendedPointId, _id),
                    vector=vector.model_dump(exclude_none=True, exclude_unset=True),
                )
            ],
            **kwargs,
        )
        return self.__is_result_success(res)

    async def aupdate_vector(
        self,
        client: AsyncQdrantClient,
        *,
        _id: qdrant_types.PointId,
        vector: VectorModelType,
        **kwargs,
    ) -> bool:
        res = await client.update_vectors(
            collection_name=self.collection,
            points=[
                qdrant_models.PointVectors(
                    id=cast(qdrant_models.ExtendedPointId, _id),
                    vector=vector.model_dump(exclude_none=True, exclude_unset=True),
                )
            ],
            **kwargs,
        )
        return self.__is_result_success(res)



from typing import Generic, List, Optional, Type, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel
from sqlalchemy import func
from sqlalchemy.orm.session import Session as SaSession
from sqlmodel import Session, select

from ..schemas import QueryBase, Uuid
from .sql_util import parse_query_for_first, parse_query_for_list

ModelType = TypeVar("ModelType", bound=BaseModel)

SqlIdType = str | None | Uuid | UUID


class SqlCRUDBase(Generic[ModelType]):
    def __init__(self, model: Type[ModelType], table_name: str):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD) in SQL

        **Parameters**
        * `model`: A SQLAlchemy model class
        """
        self.model = model
        self.model.__table__.name = table_name  # type: ignore
        self.model.__tablename__ = table_name  # type: ignore
        self.table_name = table_name

    def first(
        self, s: Session | SaSession, query: Optional[QueryBase] = None
    ) -> Optional[ModelType]:
        parsed = parse_query_for_first(self.model, query=query)
        if query is not None:
            sel = select(self.model).order_by(query.order_by)
        else:
            sel = select(self.model)
        statement = sel.limit(1)  # only the first one
        if parsed.where is not None:
            statement = statement.where(parsed.where)
        results = s.exec(statement)  # type: ignore
        return results.one_or_none()

    def get_multi(
        self, s: Session | SaSession, *, query: QueryBase | dict | None
    ) -> List[ModelType]:
        if isinstance(query, dict):
            query = QueryBase(**query)
        # assert query.offset is not None and query.limit is not None, '[Db] offset and limit is require for listing query'
        parsed = parse_query_for_list(self.model, query=query)
        statement = select(self.model)
        if query is not None and query.order_by is not None:
            statement = statement.order_by(query.order_by)
        if query is not None and query.limit is not None:
            statement = statement.limit(query.limit)
        if query is not None and query.offset is not None:
            assert isinstance(query.offset, int), (
                f"[Db] Expected int or None type offset in sql query, but got {query.offset}"
            )
            statement = statement.offset(query.offset)
        if parsed.where is not None:
            statement = statement.where(parsed.where)
        results = s.exec(statement)  # type: ignore
        return list(results.all())

    def count(self, s: Session | SaSession, query: Optional[QueryBase] = None) -> int:
        parsed = parse_query_for_list(self.model, query=query)
        statement = select(func.count(self.model.id))  # type: ignore # only the first one
        if parsed.where is not None:
            statement = statement.where(parsed.where)
        count = s.exec(statement).one()  # type: ignore
        return count

    def create(self, s: Session | SaSession, *, obj_in: ModelType):
        obj_in = self.pre_commit_create(s, obj_in=obj_in)
        s.commit()
        s.refresh(obj_in)
        return obj_in

    def pre_commit_create(self, s: Session | SaSession, *, obj_in: ModelType):
        s.add(obj_in)
        return obj_in

    def create_multi(self, s: Session | SaSession, *, obj_in_list: List[ModelType]):
        if len(obj_in_list) == 0:
            return []
        for obj_in in obj_in_list:
            s.add(obj_in)
        s.commit()
        for obj_in in obj_in_list:
            s.refresh(obj_in)
        return obj_in_list

    def update(
        self, s: Session | SaSession, *, _id: SqlIdType, obj_in: dict
    ) -> ModelType:
        obj = self.pre_commit_update(s, _id=_id, obj_in=obj_in)
        s.commit()
        s.refresh(obj)
        return obj

    def pre_commit_update(
        self,
        s: Session | SaSession,
        *,
        _id: SqlIdType,
        obj_in: dict,
    ) -> ModelType:
        if not isinstance(obj_in, dict):
            # try to convert to dict from pydantic object
            assert hasattr(obj_in, "model_dump") and callable(obj_in.model_dump), (
                f"[Db] Expected dict or pydantic object, but got type: {type(obj_in)}"
            )
            obj_in = obj_in.model_dump(exclude={"id", "created"})

        obj = self.first_by_id(s, _id=_id)
        if obj is None:
            raise ValueError("Failed to update due to object not found")
        for k, v in obj_in.items():
            # print("k, v ", k, v)
            # eval(f"obj.{k}=v")
            if hasattr(obj, k):
                setattr(obj, k, v)
            # setattr(obj, k, v)
        s.add(obj)
        return obj

    def delete(
        self,
        s: Session | SaSession,
        *,
        _id: SqlIdType,
    ):
        obj = self.first_by_id(s, _id=_id)
        if obj is None:
            return None
        s.delete(obj)
        s.commit()

    def delete_multi(
        self, s: Session | SaSession, *, obj_in_list: List[Union[str, ModelType]]
    ):
        if len(obj_in_list) == 0:
            return
        ids = [item for item in obj_in_list if isinstance(item, str)]
        objects = [item for item in obj_in_list if not isinstance(item, str)]

        if len(objects) > 0:
            for obj in objects:
                s.delete(obj)
            s.commit()
        for _id in ids:
            self.delete(s, _id=_id)

    def first_by_id(
        self, s: Session | SaSession, *, _id: SqlIdType
    ) -> Optional[ModelType]:
        if _id is None:
            raise ValueError("[Db] _id can not be None")
        return self.first(s, query=QueryBase(id=str(_id)))



import uuid
from abc import ABC, abstractmethod
from typing import Annotated, Any, List, Optional, Type, Union, get_args, get_origin
from uuid import UUID as _UUID

from bson import Binary
from bson import ObjectId as _ObjectId
from pydantic import BaseModel, ConfigDict, GetCoreSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema
from qdrant_client.conversions.common_types import PointId


def is_optional(field):
    return get_origin(field) is Union and type(None) in get_args(field)


class ObjectIdPydanticAnnotation:
    @classmethod
    def validate_object_id(cls, v: Any, handler) -> _ObjectId:
        if isinstance(v, _ObjectId):
            return v

        s = handler(v)
        if _ObjectId.is_valid(s):
            return _ObjectId(s)
        else:
            raise ValueError("Invalid ObjectId")

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type, _handler
    ) -> core_schema.CoreSchema:
        assert source_type is _ObjectId or (
            is_optional(source_type) and _ObjectId in get_args(source_type)
        ), (
            "[Db] Internal error, expected source_type is _ObjectId or (is_optional(source_type) and _ObjectId in get_args(source_type))"
        )
        return core_schema.no_info_wrap_validator_function(
            cls.validate_object_id,
            core_schema.str_schema(),
            serialization=core_schema.to_string_ser_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema, handler) -> JsonSchemaValue:
        return handler(core_schema.str_schema())


ObjectId = Annotated[_ObjectId, ObjectIdPydanticAnnotation]


class Uuid(_UUID):
    @classmethod
    def new(cls) -> _UUID:
        return uuid.uuid4()

    @classmethod
    def validate_uuid(cls, v: Any) -> _UUID:
        if isinstance(v, _UUID):
            return v
        if isinstance(v, str):
            return _UUID(v)
        if isinstance(v, Binary):
            return v.as_uuid()
        raise ValueError(f"[Db] Unidentifieable UUID data type: {v}")

    # def to_binary(self):
    #     return Binary.from_uuid(self)

    @classmethod
    def from_uuid(cls, payload) -> Binary:
        if isinstance(payload, Binary):
            return payload
        return Binary.from_uuid(payload)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.validate_uuid,
            core_schema.uuid_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                function=cls.from_uuid
            ),
        )


# add json plain serializer to Uuid
# Uuid = Annotated[
#     Uuid, PlainSerializer(lambda x: str(x), return_type=str, when_used='json') # this will override the __get_pydantic_core_schema__ declared above.
# ]

MaybeUuid = Union[Uuid, _UUID, Binary]


def to_binary_uuid(uuid_: Union[MaybeUuid, List[MaybeUuid]]):
    if isinstance(uuid_, list):
        return [to_binary_uuid(u) for u in uuid_]
    if isinstance(uuid_, Binary):
        return uuid_
    return Binary.from_uuid(uuid_)


def serialize_object_ids_in_dict(mey_be_dict: Any):
    if not isinstance(mey_be_dict, dict):
        return
    for key, val in mey_be_dict.items():
        if isinstance(val, _ObjectId):
            mey_be_dict[key] = str(val)
        elif isinstance(val, dict):
            serialize_object_ids_in_dict(val)
        elif isinstance(val, list):
            for item in val:
                serialize_object_ids_in_dict(item)


class DbSettingsABC(ABC):
    @property
    @abstractmethod
    def qdrant_host(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def qdrant_port(self) -> Optional[int]: ...

    @property
    @abstractmethod
    def mongo_db_uri(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def redis_uri(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def sql_write_uri(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def sql_read_uri(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def sql_uri(self) -> Optional[str]: ...


class QueryBase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    offset: Optional[int | str | PointId] = None
    limit: Optional[int] = None
    filters: Optional[Union[dict, List[Any]]] = None
    order_by: Optional[Any] = None

    id: Optional[Union[ObjectId, Uuid, str]] = None
    ids: Optional[List[Union[ObjectId, Uuid, str]]] = None

    def __init__(
        self,
        offset: Optional[Union[int, str, PointId]] = None,
        limit: Optional[int] = None,
        filters: Optional[Union[dict, List[Any]]] = None,
        order_by: Optional[Any] = None,
        id: Optional[Union[ObjectId, Uuid, str]] = None,
        ids: Optional[List[Union[ObjectId, Uuid, str]]] = None,
    ):
        # Perform custom initialization logic if necessary

        # Call the Pydantic BaseModel initializer to perform the standard model validation and initialization
        super().__init__(
            offset=offset,
            limit=limit,
            filters=filters,
            order_by=order_by,
            id=id,
            ids=ids,
        )

    @classmethod
    def loop(cls, func, *args, **kwargs) -> List[Any]:
        page = 0
        page_size = 100
        res = []
        has_more = True
        base_query: QueryBase | None = kwargs.get("query")
        if "query" in kwargs:
            del kwargs["query"]
        base_query_dict: dict = (
            base_query.model_dump() if base_query is not None else {}
        )
        if "offset" in base_query_dict:
            del base_query_dict["offset"]
        if "limit" in base_query_dict:
            del base_query_dict["limit"]
        while page < 10000 and has_more:
            query = QueryBase(
                **base_query_dict, limit=page_size, offset=page * page_size
            )

            output = func(*args, **kwargs, query=query)
            if len(output) == 0:
                break
            has_more = len(output) == page_size
            res += output
            page += 1
        return res

    @classmethod
    def qdrant_scroll(cls, func, *args, **kwargs) -> List[Any]:
        page = 0
        page_size = 100
        offset = None  # point id
        res = []
        has_more = True
        base_query: QueryBase | None = kwargs.get("query")
        if "query" in kwargs:
            del kwargs["query"]

        base_query_dict: dict = (
            base_query.model_dump() if base_query is not None else {}
        )
        if "offset" in base_query_dict:
            del base_query_dict["offset"]
        if "limit" in base_query_dict:
            del base_query_dict["limit"]
        while page < 10000 and has_more:
            query = QueryBase(**base_query_dict, limit=page_size, offset=offset)
            output, next_point_id = func(*args, **kwargs, query=query)
            if len(output) == 0:
                break
            has_more = next_point_id is not None
            res += output
            page += 1
            offset = next_point_id
        return res



from .crud import *
from .deps import *
from .mongo import (
    build_mongo_find_query,
    merge_mongo_and_queries,
    merge_mongo_or_queries,
)
from .schemas import (
    DbSettingsABC,
    MaybeUuid,
    ObjectId,
    QueryBase,
    Uuid,
    serialize_object_ids_in_dict,
    to_binary_uuid,
)
from .session import Db
from .wrapper import *

# from .response_encoder import CustomJSONEncoder



from typing import Optional

from redis import Redis
from redis.asyncio import Redis as AsyncRedis


class ExtendedRedis(Redis):
    def __init__(self, *args, **kwargs):
        self._namespace: Optional[str] = None
        super().__init__(*args, **kwargs)

    @classmethod
    def extend(cls, redis: Redis, *, namespace: str | None = None):
        return ExtendedRedis._create_with_namespace(redis=redis, namespace=namespace)

    def clone_with_namespace(self, namespace: str) -> "ExtendedRedis":
        """
        Creates a new Redis instance with the same connection parameters.
        """
        return self._create_with_namespace(namespace=namespace, redis=self)

    def delete_namespace_keys(self):
        if self._namespace is None:
            raise ValueError("[Db] Namespace not set")
        # Pattern for keys to be deleted
        pattern = f"{self._namespace}:*"

        # Get keys matching the pattern
        keys = self.keys(pattern)
        # Delete the keys
        if keys:
            self.delete(*keys)  # type: ignore

    @classmethod
    def _create_with_namespace(
        cls, redis: Redis, *, namespace: str | None
    ) -> "ExtendedRedis":
        """
        Creates a new Redis instance with the same connection parameters.
        """
        # Retrieve the connection parameters from the current Redis instance
        connection_params = {
            "host": redis.connection_pool.connection_kwargs.get("host", "localhost"),
            "port": redis.connection_pool.connection_kwargs.get("port", 6379),
            "db": redis.connection_pool.connection_kwargs.get("db", 0),
            "password": redis.connection_pool.connection_kwargs.get("password", None),
            "socket_timeout": redis.connection_pool.connection_kwargs.get(
                "socket_timeout", None
            ),
        }

        # Create and return a new instance with the same connection params
        clone = ExtendedRedis(**connection_params)
        clone.set_namespace(namespace)
        return clone

    def set_namespace(self, namespace: str | None):
        self._namespace = namespace  # set namespace

    def get_namespace_key(self, key: str) -> str:
        if self._namespace is None:
            raise ValueError("[Db] Namespace not set")
        return f"{self._namespace}:{key}"


class ExtendedAsyncRedis(AsyncRedis):
    def __init__(self, *args, **kwargs):
        self._namespace: Optional[str] = None
        super().__init__(*args, **kwargs)

    @classmethod
    def extend(cls, aredis: AsyncRedis, *, namespace: str | None = None):
        return ExtendedAsyncRedis._create_with_namespace(
            aredis=aredis, namespace=namespace
        )

    def clone_with_namespace(self, namespace: str) -> "ExtendedAsyncRedis":
        """
        Creates a new Redis instance with the same connection parameters.
        """
        return self._create_with_namespace(namespace=namespace, aredis=self)

    def delete_namespace_keys(self):
        if self._namespace is None:
            raise ValueError("[Db] Namespace not set")
        # Pattern for keys to be deleted
        pattern = f"{self._namespace}:*"

        # Get keys matching the pattern
        keys = self.keys(pattern)
        # Delete the keys
        if keys:
            self.delete(*keys)  # type: ignore

    @classmethod
    def _create_with_namespace(
        cls, aredis: AsyncRedis, *, namespace: str | None
    ) -> "ExtendedAsyncRedis":
        """
        Creates a new Redis instance with the same connection parameters.
        """
        # Retrieve the connection parameters from the current Redis instance
        connection_params = {
            "host": aredis.connection_pool.connection_kwargs.get("host", "localhost"),
            "port": aredis.connection_pool.connection_kwargs.get("port", 6379),
            "db": aredis.connection_pool.connection_kwargs.get("db", 0),
            "password": aredis.connection_pool.connection_kwargs.get("password", None),
            "socket_timeout": aredis.connection_pool.connection_kwargs.get(
                "socket_timeout", None
            ),
        }

        # Create and return a new instance with the same connection params
        clone = ExtendedAsyncRedis(**connection_params)
        clone.set_namespace(namespace)
        return clone

    def set_namespace(self, namespace: str | None):
        self._namespace = namespace  # set namespace

    def get_namespace_key(self, key: str) -> str:
        if self._namespace is None:
            raise ValueError("[Db] Namespace not set")
        return f"{self._namespace}:{key}"



import logging
from typing import Dict, List, Mapping, Optional, Union

from motor.motor_asyncio import AsyncIOMotorClient
from pymongo import IndexModel, MongoClient
from qdrant_client import AsyncQdrantClient, QdrantClient
from qdrant_client.http import models as qdrant_models
from qdrant_client.http.exceptions import UnexpectedResponse
from qdrant_client.http.models import PayloadSchemaType
from redis import ConnectionPool
from redis import Redis as R
from redis.asyncio import ConnectionPool as AsyncConnectionPool
from redis.asyncio import Redis as AsyncRedis
from sqlalchemy import Engine
from sqlmodel import Session, create_engine

from keble_db.wrapper import ExtendedAsyncRedis
from keble_db.wrapper.redis import ExtendedRedis

from .schemas import DbSettingsABC

logger = logging.getLogger(__name__)


class Db:
    def __init__(self, settings: DbSettingsABC):
        self.__settings = settings
        self.__redis_pool = (
            ConnectionPool.from_url(settings.redis_uri, decode_responses=True)
            if settings.redis_uri is not None
            else None
        )
        self.__aredis_pool = (
            AsyncConnectionPool.from_url(settings.redis_uri, decode_responses=True)
            if settings.redis_uri is not None
            else None
        )
        self.__sql_read_engine = (
            create_engine(
                str(
                    settings.sql_read_uri
                    if settings.sql_read_uri is not None
                    else settings.sql_uri
                ),
                pool_pre_ping=True,
                pool_size=1000,
                max_overflow=10,
                pool_recycle=3600,
                pool_use_lifo=True,
                echo=False,
            )
            if settings.sql_read_uri is not None or settings.sql_uri is not None
            else None
        )
        self.__sql_write_engine = (
            create_engine(
                str(
                    settings.sql_write_uri
                    if settings.sql_write_uri is not None
                    else settings.sql_uri
                ),
                pool_pre_ping=True,
                pool_size=1000,
                max_overflow=10,
                pool_recycle=3600,
                pool_use_lifo=True,
                echo=False,
            )
            if settings.sql_write_uri is not None or settings.sql_uri is not None
            else None
        )

        self._sql_read_instance = None
        self._sql_write_instance = None
        self._redis_instance = None
        self._aredis_instance = None
        self._mongo_instance = None
        self._amongo_instance: Optional[AsyncIOMotorClient] = None
        self._aqdrant_instance: Optional[AsyncQdrantClient] = None
        self._qdrant_instance = None

    @property
    def sql_write_engine(self) -> Optional[Engine]:
        return self.__sql_write_engine

    @property
    def sql_read_engine(self) -> Optional[Engine]:
        return self.__sql_read_engine

    def get_redis(self, *, force_new_instance: bool = False) -> R:
        """Return conn to redis
        You do not need to explicitly close the redis conn"""
        if self.__redis_pool is None:
            raise ValueError("[Db] redis_pool required")
        if force_new_instance:
            return R(connection_pool=self.__redis_pool)
        if not force_new_instance and self._redis_instance is not None:
            return self._redis_instance
        self._redis_instance = R(connection_pool=self.__redis_pool)
        return self._redis_instance

    def get_aredis(self, *, new_instance: bool = True) -> AsyncRedis:
        if self.__aredis_pool is None:
            raise ValueError("[Db] aredis_pool required")
        if new_instance:
            return AsyncRedis(connection_pool=self.__aredis_pool, decode_responses=True)
        if not new_instance and self._aredis_instance is not None:
            return self._aredis_instance
        self._aredis_instance = AsyncRedis(
            connection_pool=self.__aredis_pool, decode_responses=True
        )
        return self._aredis_instance

    def get_extended_aredis(
        self, *, new_instance: bool = True, namespace: Optional[str] = None
    ) -> ExtendedAsyncRedis:
        return ExtendedAsyncRedis.extend(
            aredis=self.get_aredis(new_instance=new_instance), namespace=namespace
        )

    def get_extended_redis(
        self, *, force_new_instance: bool = False, namespace: Optional[str] = None
    ) -> ExtendedRedis:
        return ExtendedRedis.extend(
            redis=self.get_redis(force_new_instance=force_new_instance),
            namespace=namespace,
        )

    def get_mongo(self, *, new_instance: bool = True) -> MongoClient:
        """Return a mongo conn
        You do not need to close PyMongo connections. Leave them open so that PyMongo connection pooling gives you the most efficient performance"""
        if self.__settings.mongo_db_uri is None:
            raise ValueError("[Db] mongo_db_uri required")

        if new_instance:
            return MongoClient(
                self.__settings.mongo_db_uri, uuidRepresentation="standard"
            )

        if self._mongo_instance is not None:
            try:
                self._mongo_instance.server_info()
                return self._mongo_instance
            except Exception as e:
                logger.critical(
                    f"[Db] Old mongo instance has been closed, now create a new one and returned. Server info exception: {e}"
                )
                pass

        self._mongo_instance = MongoClient(
            self.__settings.mongo_db_uri, uuidRepresentation="standard"
        )
        return self._mongo_instance

    async def aget_amongo(self, *, new_instance: bool = True) -> AsyncIOMotorClient:
        """Return a async mongo conn
        You do not need to close PyMongo connections. Leave them open so that PyMongo connection pooling gives you the most efficient performance"""
        if self.__settings.mongo_db_uri is None:
            raise ValueError("[Db] mongo_db_uri required")

        if new_instance:
            # check new instance first, new instance is independent and will not save as self.instance
            return AsyncIOMotorClient(
                self.__settings.mongo_db_uri, uuidRepresentation="standard"
            )

        if self._amongo_instance is not None:
            try:
                await self._amongo_instance.server_info()
                return self._amongo_instance
            except Exception as e:
                logger.critical(
                    f"[Db] Old async mongo instance has been closed, now create a new one and returned. Server info exception: {e}"
                )
                pass

        self._amongo_instance = AsyncIOMotorClient(
            self.__settings.mongo_db_uri, uuidRepresentation="standard"
        )
        return self._amongo_instance

    def get_amongo(self) -> AsyncIOMotorClient:
        """Return a async mongo conn
        without checking old instance, use with caustious to prevent creating too many session/conn to mongo at the same time.
        """
        if self.__settings.mongo_db_uri is None:
            raise ValueError("[Db] mongo_db_uri required")

        return AsyncIOMotorClient(
            self.__settings.mongo_db_uri, uuidRepresentation="standard"
        )

    def get_qdrant_client(self, *, new_instance: bool = True) -> QdrantClient:
        if self.__settings.qdrant_host is None or self.__settings.qdrant_port is None:
            raise ValueError("[Db] qdrant_host and qdrant_port required")
        if new_instance:
            # check new instance first, new instance is independent and will not save as self.instance
            return QdrantClient(
                host=self.__settings.qdrant_host,
                port=self.__settings.qdrant_port,
                timeout=120,
            )
        if not new_instance and self._qdrant_instance is not None:
            return self._qdrant_instance
        self._qdrant_instance = QdrantClient(
            host=self.__settings.qdrant_host,
            port=self.__settings.qdrant_port,
            timeout=120,
        )
        return self._qdrant_instance

    def get_aqdrant_client(self, *, new_instance: bool = True) -> AsyncQdrantClient:
        if self.__settings.qdrant_host is None or self.__settings.qdrant_port is None:
            raise ValueError("[Db] qdrant_host and qdrant_port required")
        if new_instance:
            # check new instance first, new instance is independent and will not save as self.instance
            return AsyncQdrantClient(
                host=self.__settings.qdrant_host,
                port=self.__settings.qdrant_port,
                timeout=120,
            )
        if not new_instance and self._aqdrant_instance is not None:
            return self._aqdrant_instance
        self._aqdrant_instance = AsyncQdrantClient(
            host=self.__settings.qdrant_host,
            port=self.__settings.qdrant_port,
            timeout=120,
        )
        return self._aqdrant_instance

    def get_sql_write_client(self, *, new_instance: bool = True) -> Session:
        if self.sql_write_engine is None:
            raise ValueError("[Db] sql_write_engine required")
        if new_instance:
            return Session(self.sql_write_engine)
        if not new_instance and self._sql_write_instance is not None:
            return self._sql_write_instance
        self._sql_write_instance = Session(self.sql_write_engine)
        return self._sql_write_instance

    def get_sql_read_client(self, *, new_instance: bool = True) -> Session:
        if self.sql_read_engine is None:
            raise ValueError("[Db] sql_read_engine required")
        if new_instance:
            return Session(self.sql_read_engine)
        if not new_instance and self._sql_read_instance is not None:
            return self._sql_read_instance
        self._sql_read_instance = Session(self.sql_read_engine)
        return self._sql_read_instance

    @staticmethod
    def ready_mongo(client: MongoClient):
        # check by list
        client.list_databases()

    @staticmethod
    def ready_mongo_indexes(
        client: MongoClient,
        db_name: str,
        collection_name: str,
        indexes: List[Dict[str, Union[int, str]]],
    ) -> None:
        """
        Ensure MongoDB indexes exist. Creates them if not.

        :param client: MongoClient instance
        :param db_name: Name of the MongoDB database
        :param collection_name: Name of the MongoDB collection
        :param indexes: List of dictionaries defining the index fields and their types.
                        - For normal ascending/descending indexes, use 1 (ascending) or -1 (descending).
                        - For text search indexes, use "text".
                        Example: [{"field1": 1, "field2": -1}, {"field3": "text"}]
        """
        collection = client[db_name][collection_name]
        existing_indexes = collection.index_information()

        # Create index models based on the input
        indexes_need_to_create = []
        for index in indexes:
            compound_index_key = [(field, order) for field, order in index.items()]

            # Check if the index already exists in the collection
            if not any(
                compound_index_key == idx.get("key")
                for idx in existing_indexes.values()
            ):
                index_model = IndexModel(compound_index_key)
                indexes_need_to_create.append(index_model)
        if len(indexes_need_to_create) > 0:
            collection.create_indexes(indexes_need_to_create)

    @staticmethod
    def ready_redis(redis: R):
        # check by get key
        redis.get("k")

    @staticmethod
    def ready_qdrant(
        client: QdrantClient,
        collection: str,
        vectors_config: Union[
            qdrant_models.VectorParams, Mapping[str, qdrant_models.VectorParams]
        ],
    ):
        # need to check database exist or not
        try:
            client.get_collection(collection_name=collection)
        except (AssertionError, UnexpectedResponse):
            # collection not found, or connection error. Try to create
            client.create_collection(
                collection_name=collection,
                vectors_config=vectors_config,
            )

    @staticmethod
    def ready_qdrant_indexes(
        client: QdrantClient,
        collection_name: str,
        indexes: List[Dict[str, PayloadSchemaType]],
    ) -> None:
        """
        Ensure Qdrant payload indexes exist. Creates them if not.

        :param client: QdrantClient instance
        :param collection_name: Name of the Qdrant collection
        :param indexes: List of dictionaries where keys are field names and values are their corresponding types (PayloadSchemaType).
                        Example: [{"field1": PayloadSchemaType.INTEGER}, {"field2": PayloadSchemaType.KEYWORD}]
        """
        try:
            collection_info = client.get_collection(collection_name)
            existing_indexes = collection_info.payload_schema or {}

            for index in indexes:
                for field_name, schema_type in index.items():
                    if field_name not in existing_indexes:
                        client.create_payload_index(
                            collection_name=collection_name,
                            field_name=field_name,
                            field_schema=qdrant_models.PayloadSchemaType(schema_type),
                        )
        except (AssertionError, UnexpectedResponse):
            logger.critical(
                f"[Db] Collection {collection_name} does not exist or failed to retrieve collection info."
            )

    @staticmethod
    def try_close(
        *sessions: Optional[
            MongoClient
            | R
            | Session
            | QdrantClient
            | AsyncIOMotorClient
            | AsyncQdrantClient
            | AsyncRedis
        ],
    ):
        for session in sessions:
            if session is not None:
                try:
                    session.close()
                except Exception as e:
                    logger.error(f"[Db] Failed to close a db session due to {e}")

    # @staticmethod
    # def try_close(
    #     *sessions: Optional[
    #         MongoClient
    #         | R
    #         | Session
    #         | QdrantClient
    #         | AsyncIOMotorClient
    #         | AsyncQdrantClient
    #         | AsyncRedis  # Keeping it broad for various session types
    #     ],
    # ):
    #     async def close_session(session):
    #         """Handles both async and sync close methods."""
    #         if session is None:
    #             return
    #         try:
    #             # close_method = getattr(session, "close", None)
    #             # if close_method is not None:
    #             #     if asyncio.iscoroutinefunction(close_method):
    #             #         await close_method()  # Await if it's an async function
    #             #     else:
    #             #         close_method()  # Call directly if it's sync
    #             close_method = getattr(session, "close", None)
    #             if close_method is not None:
    #                 if asyncio.iscoroutinefunction(close_method):
    #                     result = await close_method()  # Await if it's an async function
    #                 else:
    #                     result = close_method()  # Call directly if it's sync
    #                 # sometime asyncio.iscoroutinefunction failed to detect await, we will use result detect future
    #                 if isinstance(result, Awaitable) or asyncio.isfuture(result):
    #                     await result  # Await if it's an awaitable (coroutine/future)
    #         except Exception as e:
    #             logger.error(f"[Db] Failed to close a db session due to {e}")

    #     async def close_all():
    #         """Runs all close operations concurrently."""
    #         await asyncio.gather(*(close_session(session) for session in sessions))

    #     # Check if inside an async context, otherwise run synchronously
    #     try:
    #         if asyncio.get_running_loop().is_running():
    #             asyncio.create_task(close_all())  # Schedule without blocking
    #         else:
    #             asyncio.run(close_all())  # Run if no event loop is active
    #     except RuntimeError:  # No running loop
    #         asyncio.run(close_all())
